xquery version "3.1";
(:~
 : The TextGrid Client offers interfaces to the main features of TextGrid,
 : when they are available via REST.
 : @version 1.1
 :)

module namespace tgclient="https://sade.textgrid.de/ns/tgclient";

declare namespace http="http://expath.org/ns/http-client";
declare namespace rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#";
declare namespace sparql-results="http://www.w3.org/2005/sparql-results#";
declare namespace tgmd="http://textgrid.info/namespaces/metadata/core/2010";

(:~
 : @param $id - the TextGrid ID of the object, e.g. textgrid:1234.5
 :)

declare function tgclient:get-metadata($id as xs:string,
    $tgcrud-url as xs:string,
    $sid as xs:string?)
as node() {
    let $request-url := tgclient:make-metadata-request-URL($id, $tgcrud-url, $sid)
    let $result := tgclient:send-metadata-request($request-url, $id)
    return
        if( count($result) != 1 ) then
            <error> { $id } </error>
        else
            $result
};



declare function tgclient:make-metadata-request-URL($id as xs:string,
    $tgcrud-url as xs:string,
    $sid as xs:string?)
as xs:string {
    $tgcrud-url || "/" || $id || "/metadata?sessionId=" || $sid
};


declare function tgclient:send-metadata-request($request-url as xs:string,
    $id as xs:string)
as node() {
    let $result := tgclient:make-request($request-url)
    return
        if ($result//tgmd:MetadataContainerType) then
            $result//tgmd:MetadataContainerType
        else
            <error>URI { $id } did not yield a result.</error>
};


declare function tgclient:get-data($id as xs:string,
    $tgcrud-url as xs:string,
    $sid as xs:string?)
as item() {
    let $request-url := tgclient:make-data-request-URL($id, $tgcrud-url, $sid)
    let $request := tgclient:make-request($request-url)
    let $getBody := $request[2]

    return
        $getBody
};


declare function tgclient:make-data-request-URL($id as xs:string,
    $tgcrud-url as xs:string,
    $sid as xs:string?)
as xs:string {
    $tgcrud-url || "/" || $id || "/data?sessionId=" || $sid
};


declare function tgclient:make-request($request-url as xs:string)
as item()+ {
    let $header :=
      <http:request method="get">
        <http:header name="Connection" value="close" />
      </http:request>
    return
        try {
            http:send-request($header, $request-url)
        } catch * {
            error(QName("https://sade.textgrid.de/ns/tgclient", "TGCLIENT01"), "No object could be fetched from " || $request-url || ".")
        }
};

(:~ Returns a list of the latest TextGrid items within a given aggregation.
 : @param tguri – the URI of any tg.aggregation
 : @param rdfstore – url to the triple store
 : @return sequence of strings (textgrid URIs) or empty sequence
 :)
declare function tgclient:get-latest-aggregated-uris($tguri as xs:string,
    $rdfstore as xs:string)
as xs:string* {
    let $sparql-result := tgclient:get-sparql-result($tguri, $rdfstore)
    let $uris := tgclient:get-unrevisioned-aggregated-uris($sparql-result)
    for $uri in $uris
        let $max-rev := tgclient:find-max-rev($sparql-result, $uri)
        return
            $uri || '.' || $max-rev
};


declare function tgclient:get-sparql-result($tguri as xs:string,
    $rdfstore as xs:string)
as element(sparql-results:sparql) {
    let $query := tgclient:make-sparql-query-for-aggregated-uris($tguri)
    return
        tgclient:sparql($query, $rdfstore)
};


declare function tgclient:make-sparql-query-for-aggregated-uris($tguri as xs:string)
as xs:string {
    concat("
    PREFIX ore:<http://www.openarchives.org/ore/terms/>
    PREFIX tg:<http://textgrid.info/relation-ns#>
    
    SELECT ?s
        WHERE { <",$tguri,"> (ore:aggregates/tg:isBaseUriOf|ore:aggregates)* ?s.
        FILTER not exists { ?s tg:isDeleted true } . }")
    => normalize-space()
};


declare function tgclient:get-unrevisioned-aggregated-uris($sparql-result as element(sparql-results:sparql))
as xs:string+ {
    distinct-values($sparql-result//sparql-results:uri[not(. = "")]/substring-before(.,'.'))
};


declare function tgclient:find-max-rev($sparql-result as element(sparql-results:sparql),
    $current-uri as xs:string)
as xs:integer {
    let $versions := $sparql-result//sparql-results:uri[starts-with(., $current-uri)]
    let $revisions := $versions/xs:integer(substring-after(., '.'))
    return
        max($revisions)
};


(:
 : Queries TextGrid RDF store. Mainly used to get URIs within a collection
 : @param $query - the SPARQL query as string
 : @param $rdfstore - URL of the public SPARQL endpoint
 : @return the result of the SPARQL query as XML node
 :)
declare function tgclient:sparql($query as xs:string,
    $rdfstore as xs:string)
as node() {
    let $reqUrl := tgclient:get-sparql-request-url($query, $rdfstore)
    let $header :=
      <http:request method="get">
        <http:header
          name = "accept"
          value = "text/html,application/xhtml+xml,application/xml;q=0.9,*/*;q=0.8"
        />
        <http:header
          name = "connection"
          value = "close"
        />
      </http:request>

    return http:send-request($header, $reqUrl)[2]/node()
};



declare function tgclient:get-sparql-request-url($query as xs:string,
    $rdfstore as xs:string)
as xs:string {
    let $urlEncodedQuery as xs:string := encode-for-uri($query)
    return
        string-join(($rdfstore, "?query=", $urlEncodedQuery))
};



declare function tgclient:remove-prefix($tguri as xs:string) as xs:string {
    if (contains($tguri, ":")) then
        tokenize($tguri, ":")[position() gt 1] => string-join()
    else
        $tguri
};

