xquery version "3.1";

(:~
 : This module provides the functions to transfer data from TextGrid to this
 : instance. It will be called via JavaScript XHR requests from index.html.
 :)

module namespace tgconnect="https://sade.textgrid.de/ns/connect";

import module namespace config="http://sub.uni-goettingen.de/tg-connect/config" at "config.xqm";
import module namespace tgclient="https://sade.textgrid.de/ns/tgclient" at "client.xqm";

declare namespace tei="http://www.tei-c.org/ns/1.0";
declare namespace tgmd="http://textgrid.info/namespaces/metadata/core/2010";


(:~ The main function for the publisher.
 : @param $uri – the TextGrid URI of a single object
 : @param $sid – a valid SessionId able to get the objects
 : @param $user – a username for the db
 : @param $password – the password for the user in the db
 :)
declare function tgconnect:publish($uri as xs:string,
    $sid as xs:string,
    $user as xs:string,
    $password as xs:string)
{
    if (tgconnect:is-user-authorized($user, $password)) then
        let $tgcrud-url := config:get-value-from-configfile("textgrid.tgcrud")
        let $metadataContainer := tgclient:get-metadata($uri, $tgcrud-url, $sid)
        let $tguri := $metadataContainer//tgmd:textgridUri/string()
        let $rdfstoreUrl := tgconnect:get-triplestore-url($metadataContainer)
        
        return
        for $public-uri in tgclient:get-latest-aggregated-uris($tguri, $rdfstoreUrl)
            let $data-file := tgclient:get-data($public-uri, $tgcrud-url, $sid)
            let $meta := tgconnect:get-metadata($public-uri, $tguri, $metadataContainer, $tgcrud-url, $sid)
            let $targetUri := tgconnect:make-target-uri($meta)
            let $format := string($meta//tgmd:format)
            return
                (
                    tgconnect:process-according-to-format($format, $meta, $targetUri, $data-file),
                    tgconnect:store-metadata($format, $targetUri, $meta)
                )

    else
        error(
            QName("https://sade.textgrid.de/ns/error", "PUBLISH02"),
            "error authenticating for " || $user || " on " || $config:data-root
            )
};

(: we should not download the metadata twice :)
declare function tgconnect:get-metadata($public-uri as xs:string,
    $tguri as xs:string,
    $metadataContainer as element(tgmd:MetadataContainerType),
    $tgcrud-url as xs:string,
    $sid as xs:string)
as element(tgmd:MetadataContainerType) {
    if($public-uri = $tguri) then
        $metadataContainer
    else
        tgclient:get-metadata($public-uri, $tgcrud-url, $sid)
};



declare function tgconnect:store-metadata($format as xs:string,
    $targetUri as xs:string,
    $meta as element(tgmd:MetadataContainerType))
as xs:string? {
    if (tgconnect:is-image($format)) then
        ()
    else
        xmldb:store($config:data-root || "/meta", $targetUri, $meta, "text/xml")
};


declare function tgconnect:process-according-to-format($format as xs:string,
    $meta as element(tgmd:MetadataContainerType),
    $targetUri as xs:string,
    $data-file as item())
as item()+ {
    if (tgconnect:is-image($format)) then
        tgconnect:store-image($meta)
        
    else if (tgconnect:is-no-further-processing-necessary($format)) then
        tgconnect:store($data-file, $format, $targetUri)
    
    else if ($format = "text/xml") then
        tgconnect:process-xml($data-file, $targetUri)
    
    else if ($format = "text/xsd+xml") then
        tgconnect:process-xsd($data-file, $format, $targetUri)
    
    else if ($format = "text/plain") then
        tgconnect:process-plain-text($data-file, $targetUri, $meta)
    
    else
        "The publisher does not know how to handle " || $format || "."
};


declare function tgconnect:is-no-further-processing-necessary($format as xs:string)
as xs:boolean {
    if (tgconnect:is-aggregation($format)
    or $format = "text/xml+xslt"
    or $format = "text/tg.inputform+rdf+xml"
    or $format = "text/linkeditorlinkedfile") then
        true()
    else
        false()
};


declare function tgconnect:process-xsd($data-file as item(),
    $format as xs:string,
    $targetUri as xs:string)
as xs:string {
    let $targetUri := replace($targetUri, "xml", "xsd.xml")
    return
        tgconnect:store($data-file, $format, $targetUri)
};


declare function tgconnect:store-xml($data as item(),
    $targetUri as xs:string)
as xs:string? {
    let $name := tgconnect:make-xml-filename($data, $targetUri)
    return
        xmldb:store($config:data-root||"/data", $name, $data, "text/xml")
};


declare function tgconnect:make-xml-filename($data as node(),
    $targetUri as xs:string)
as xs:string {
    if($data/*/namespace-uri() = "http://relaxng.org/ns/structure/1.0") then
        $targetUri || ".rng"
    else
        $targetUri
};


declare function tgconnect:process-xml($data as item(),
    $targetUri as xs:string)
as xs:string {
    if (tgconnect:is-xml-valid($data)) then
        tgconnect:store-xml($data, $targetUri)
    else
        error(QName("https://sade.textgrid.de/ns/connect", "PUBLISH01"), 
       "The XML file with the URI " || $targetUri || " is invalid and cannot be stored.")
};


declare function tgconnect:is-xml-valid($data as item())
as xs:boolean {
    let $relaxNGPath :=  tgconnect:get-relaxNG-path($data)
    return
        if(tgconnect:is-validatable($relaxNGPath)) then
             validation:jing($data, doc($relaxNGPath))
        else
           error(QName("https://sade.textgrid.de/ns/connect", "PUBLISH02"), 
           "No RelaxNG schema found in the database for validating this XML file.")
};

declare function tgconnect:is-validatable($relaxNGPath as xs:string)
as xs:boolean {
    doc-available($relaxNGPath)
};


declare function tgconnect:get-relaxNG-path($data as item())
as xs:string {
    let $relaxNG-uri := $data/processing-instruction()
                        [contains(string(.), "http://relaxng.org/ns/structure/1.0")]
                        /substring-after(substring-before(., '" type'), "textgrid:")
    return
        $config:data-root || "/data/" || $relaxNG-uri || ".xml.rng"
};


declare function tgconnect:process-plain-text($data-file as item(),
    $targetUri as xs:string,
    $metadata-file as element(tgmd:MetadataContainerType))
as xs:string {
    let $name := replace($targetUri, "xml", "txt")
    return
        if(tgconnect:is-lucene-config($metadata-file)) then
            tgconnect:process-lucene-config-files($data-file, $metadata-file)
        else
            xmldb:store-as-binary($config:data-root || "/data", $name, $data-file)
};


declare function tgconnect:is-lucene-config($metadata-file as element(tgmd:MetadataContainerType))
as xs:boolean {
    tgconnect:get-title($metadata-file) = ("synonyms.txt", "charmap.txt")
};



declare function tgconnect:process-lucene-config-files($text as item(),
    $metadata-file as element(tgmd:MetadataContainerType))
as xs:string {
    let $title := tgconnect:get-title($metadata-file)
    let $base64 := xs:base64Binary(util:base64-encode($text))
    let $path := system:get-exist-home() || util:system-property("file.separator")
    let $store-on-disk :=
        try     { file:serialize-binary($base64, $path || $title) }
        catch * { error(
                    QName("https://sade.textgrid.de/ns/error", "PUBLISH07"),
                    "storing Lucene configuration failed with " || $err:code || ": " || $err:description)
                }
    let $reindex := xmldb:reindex( $config:data-root || "/data" )
    return
        $path || $title
};


declare function tgconnect:get-title($metadata-file as element(tgmd:MetadataContainerType))
as xs:string {
    $metadata-file//tgmd:title
};


(: we do not store images, but reference the published ones in the file images.xml :)
declare function tgconnect:store-image($metadata-file as element(tgmd:MetadataContainerType)) {
    if(doc-available( $config:data-root || "/images.xml" )) then
        tgconnect:update-image-xml($metadata-file)
    else
        tgconnect:create-and-update-image-xml($metadata-file),
    "Image: some metadata stored."
};


declare function tgconnect:create-and-update-image-xml($metadata-file as element(tgmd:MetadataContainerType)) {
    xmldb:store($config:data-root, "images.xml", <images/>),
    tgconnect:update-image-xml($metadata-file)
};


declare function tgconnect:update-image-xml($metadata-file as element(tgmd:MetadataContainerType)) {
    update insert
        <image
            uri="{string($metadata-file//tgmd:textgridUri)}"
            title="{string($metadata-file//tgmd:title)}"
            format="{string($metadata-file//tgmd:format)}"
        />
    into doc($config:data-root || "/images.xml")/images
};


declare function tgconnect:is-user-authorized($user as xs:string,
    $password as xs:string)
as xs:boolean {
    xmldb:login($config:app-root, $user, $password)
};


declare function tgconnect:make-request-url-for-data($tgcrud-url as xs:string,
    $public-uri as xs:string,
    $sid as xs:string?)
as xs:string {
    $tgcrud-url || "/" || $public-uri || "/data?sessionId=" || $sid
};


declare function tgconnect:make-request-url-for-metadata($tgcrud-url as xs:string,
    $public-uri as xs:string,
    $sid as xs:string?)
as xs:string {
    $tgcrud-url || "/" || $public-uri || "/metadata?sessionId=" || $sid
};



declare function tgconnect:make-target-uri($metadata-file as element(tgmd:MetadataContainerType)) {
    let $preserve-revisions as xs:boolean := config:get-value-from-configfile("textgrid.preserveRevisions") => xs:boolean()
    return
        tgclient:remove-prefix($metadata-file//tgmd:textgridUri/text())
        !  (if($preserve-revisions)
            then string(.)
            else substring-before(., '.'))
        || ".xml"
};


declare function tgconnect:get-triplestore-url($metadata-file as element(tgmd:MetadataContainerType))
as xs:string {
    if ($metadata-file//tgmd:generated/tgmd:availability = "public") then
        config:get-value-from-configfile("textgrid.public-triplestore")
    else
        config:get-value-from-configfile("textgrid.nonpublic-triplestore")
};


declare function tgconnect:is-image($format as xs:string)
as xs:boolean {
    starts-with($format, "image")
};


declare function tgconnect:is-aggregation($format as xs:string)
as xs:boolean {
    contains($format, "tg.aggregation")
};


declare function tgconnect:determine-target-collection($format as xs:string)
as xs:string {
    if (tgconnect:is-aggregation($format)) then
        "agg"
    else if ($format = "text/xml+xslt"
    or $format = "text/xsd+xml") then
        "data"
    else if ($format = "text/tg.inputform+rdf+xml") then
        "rdf"
    else if ($format = "text/linkeditorlinkedfile") then
        "tile"
    else
        "data"
};


declare function tgconnect:store($data as item(),
    $format as xs:string,
    $uri as xs:string)
as xs:string {
    let $subcollection := tgconnect:determine-target-collection($format)
    let $target-collection := $config:data-root || '/' || $subcollection
    return
        xmldb:store($target-collection, $uri, $data, "text/xml")
};